
#ifndef SSE2_EMULATE_INCLUDED
#define SSE2_EMULATE_INCLUDED 1

#ifndef __SSE2__
#define __SSE2__ 1

#include <xmmintrin.h>
typedef __m128 __m128i;
typedef __m128 __m128d;

union sse4 {
	__m128 m;
	__m128i mi;
	__m128d md;
	__m64  h[2];
	double d[2];
	float  f[4];
	int    i[4];
};

inline   void _mm_store_si128(__m128i *a, __m128i b)
{
	_mm_store_ps((float*) a, b);
}

inline __m128i _mm_and_si128( __m128i a, __m128i b)
{
	return _mm_and_ps(a, b);
}

inline __m128i _mm_andnot_si128( __m128i mask, __m128i a)
{
	return _mm_andnot_ps( mask, a );
}

inline __m128i _mm_or_si128( __m128i a, __m128i b)
{
	return _mm_or_ps(a, b);
}

#define _mm_shuffle_epi32( X, Y)  _mm_shuffle_ps( (X),(X),(Y))

//inline __m128i _mm_shuffle_epi32( __m128i a, int imm)
//{
//	return _mm_shuffle_ps( a, a, imm );  // compiler doesn't like this
//}

inline __m128i _mm_set1_epi32( int a )
{
	sse4 t;    t.i[0] = a;    return _mm_set_ps1(t.f[0]);
}

inline __m128i _mm_set_epi32( int d, int c, int b, int a )
{
	sse4 t;
	t.i[0] = a; t.i[1] = b; t.i[2] = c; t.i[3] = d;
	return (__m128i)_mm_load_ps(&t.f[0]);
}

inline __m128d _mm_set1_pd( double a )
{
	sse4 t;    t.d[0] = t.d[1] = a; return t.md;
}

inline __m128d _mm_set_pd( double b, double a )
{
	sse4 t;	   t.d[0] = a; t.d[1] = b;  return t.md;
}

inline __m128d _mm_add_pd( __m128d a, __m128d b )
{
#if 0
	union sse4 aa = {a};
	union sse4 bb = {b};
	sse4 t;
	t.d[0] = aa.d[0]+bb.d[0];
	t.d[1] = aa.d[1]+bb.d[1];
	return t.md;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
	aa.d[0]+=bb.d[0];
	aa.d[1]+=bb.d[1];
	return aa.md;
#endif
}

inline __m128d _mm_sub_pd( __m128d a, __m128d b )
{
#if 0
	union sse4 aa = {a};
	union sse4 bb = {b};
	sse4 t;
	t.d[0] = aa.d[0]-bb.d[0];
	t.d[1] = aa.d[1]-bb.d[1];
	return t.md;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
	aa.d[0] -= bb.d[0];
	aa.d[1] -= bb.d[1];
	return aa.md;
#endif
}

inline __m128d _mm_mul_pd( __m128d a, __m128d b )
{
#if 0
	union sse4 aa = {a};
	union sse4 bb = {b};
	sse4 t;
	t.d[0] = aa.d[0]*bb.d[0];
	t.d[1] = aa.d[1]*bb.d[1];
	return t.md;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
	aa.d[0] *= bb.d[0];
	aa.d[1] *= bb.d[1];
	return aa.md;
#endif
}

inline   void _mm_store_pd( double *a, __m128d b)
{
	_mm_store_ps((float*) a, b);
}

inline __m128i _mm_add_epi32( __m128i a, __m128i b)
{
#if 0
	sse4 t;
	union sse4 aa = {a};
	union sse4 bb = {b};
	t.h[0] = _mm_add_pi32( aa.h[0], bb.h[0]);
	t.h[1] = _mm_add_pi32( aa.h[1], bb.h[1]);
    _mm_empty();
	return t.mi;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
//	aa.h[0] = _mm_add_pi32( aa.h[0], bb.h[0]);
	aa.i[0] += bb.i[0];
	aa.i[1] += bb.i[1];
	aa.i[2] += bb.i[2];
	aa.i[3] += bb.i[3];	
//	aa.h[1] = _mm_add_pi32( aa.h[1], bb.h[1]);
//    _mm_empty();
	return aa.mi;
#endif
}

inline __m128i _mm_cmplt_epi32 (__m128i a, __m128i b)
{
#if 0
	sse4 t;
	union sse4 aa = {a};
	union sse4 bb = {b};
	t.h[0] = _mm_cmpgt_pi32( bb.h[0], aa.h[0] );
	t.h[1] = _mm_cmpgt_pi32( bb.h[1], aa.h[1] );
	_mm_empty();
	return t.m;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
	aa.h[0] = _mm_cmpgt_pi32( bb.h[0], aa.h[0] );
	aa.h[1] = _mm_cmpgt_pi32( bb.h[1], aa.h[1] );
	_mm_empty();
	return aa.mi;
#endif
}

inline __m128i _mm_cmpgt_epi32 (__m128i a, __m128i b)
{
#if 0
	sse4 t;
	union sse4 aa = {a};
	union sse4 bb = {b};
	t.h[0] = _mm_cmpgt_pi32( aa.h[0], bb.h[0] );
	t.h[1] = _mm_cmpgt_pi32( aa.h[1], bb.h[1] );
	_mm_empty();
	return t.m;
#else
	union sse4 aa = {a};
	union sse4 bb = {b};
	aa.h[0] = _mm_cmpgt_pi32( aa.h[0], bb.h[0] );
	aa.h[1] = _mm_cmpgt_pi32( aa.h[1], bb.h[1] );
	_mm_empty();
	return aa.m;
#endif
}

inline __m128 _mm_cvtpd_ps (__m128d a)
{
#if 0
	sse4 t;
	union sse4 aa = {a};
	t.f[0] = (float)aa.d[0];
	t.f[1] = (float)aa.d[1];
	t.f[2] = t.f[3] = 0.0f;
	return t.m;
#else
	union sse4 aa = {a};
	aa.f[0] = (float)aa.d[0];
	aa.f[1] = (float)aa.d[1];
	aa.f[2] = aa.f[3] = 0.0f;
	return aa.m;
#endif
}

inline int _mm_cvtsi128_si32 (__m128i a)
{
	union sse4 aa = {a};
	return aa.i[0];
}

inline __m128i _mm_cvtsi32_si128(int a)
{
	union sse4 b = { 0x0 };
	b.i[0] = a;
	return _mm_load_ps(b.f);
}

inline __m128i _mm_setzero_si128()
{
	return _mm_set1_epi32(0x0);
}

// needed for assy rounding routine
#define FPUROUND 6755399441055744.0 * 2048.0
// =============================================================================
// macros to help out with different compilers
// =============================================================================


// ALIGNED_YES Macro - tell	compiler buffer	is already SIMD	aligned,
// optimize	appropriately
//

#define SIMD_ALIGN 16

	// Make variable lie on correct boundary
#ifdef __GNUC__
#define ALIGN_SIMD __attribute__ ((aligned(SIMD_ALIGN)))
#else
#define ALIGN_SIMD __declspec(align(SIMD_ALIGN))
#endif

	// Tell compiler "pointer points to aligned data"
#if defined( __INTEL_COMPILER )
	#define	ALIGNED_YES( buffer_ )	__assume_aligned( buffer_, SIMD_ALIGN );
#else
	#define	ALIGNED_YES( buffer_ )
#endif


#endif    /* _SSE2_ */
#endif    /* SSE2_EMULATE_INCLUDED  */

