#include "astropulse.h"
#ifdef BOINC_APP_GRAPHICS
  #include "ap_graphics.h"
#endif
#if USE_FFTW
	#include "fftw3.h"
#endif
#include "sbtf.h"
#include "ap_debug.h"
#include <stdio.h>
#include "mtrand.h"
#include <complex>
#if USE_ACML
	#include "acml.h"
#endif
//using std::complex;
//#include "seti_dr2filter.h"

namespace Astropulse {

void Science::print_DC_strengths() {

  const int NUM_STRENGTHS = 8;
  float strengths[NUM_STRENGTHS];
  float totals[NUM_STRENGTHS];
  float averages[NUM_STRENGTHS];
  float maxima[NUM_STRENGTHS];
  // long fft_len = client.ap_shmem->ap_gdata.state.fft_len;
  int fft_len2 = 1024;
  // int fft_len2 = 32768;
  long datasize = 8 * 1024 * 1024;
  FILE* outfile = fopen("temp.txt", "w");
  int num_ffts = 0;

  for(int i = 0; i < NUM_STRENGTHS; i++) {
	totals[i] = 0;
	maxima[i] = 0;
  }

  // Note rawdata is in BYTES, each byte is 4 samples, fft_len is in samples, so # bytes = fft_len2 / 4
  // data_chunk_now is also in bytes, but it INCREMENTS by fft_len2 /2 / 4 to produce overlap.
  // That is, it only increments by HALF of a FFT length each iteration.
  // To find # of samples from data_chunk_now, MULTIPLY BY 4!
  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
	num_ffts++;
#if (__SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *)i_data, fft_len2))

/*#elif _GNU_SOURCE && (__SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *)i_data, fft_len2))*/

#else
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now],(float *)data, fft_len2))
#endif
	{
	  exit(SBTF_ERR);
	}
#if USE_FFTW
	fftwf_execute(fp2); // Note that this fft knows about fft_len2, and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024
#elif USE_ACML
	cfft1d(-1,acmlfftlen2,(complex*)data,fp2,&info);
#endif
	for(int i = 0; i < NUM_STRENGTHS; i++) {
	  strengths[i] = 0;
	  long amt_to_subtract = (i==0)?0:pow_int(2,i-1);
	  for(long j = 0; j < pow_int(2,i); j++) {
		float real = ((ap_complex *) data)[(j - amt_to_subtract + fft_len2) % fft_len2][0];
		float imag = ((ap_complex *) data)[(j - amt_to_subtract + fft_len2) % fft_len2][1];
		strengths[i] += real * real + imag * imag;
	  }
	  totals[i] += strengths[i];
	  maxima[i] = (strengths[i] > maxima[i])?strengths[i]:maxima[i];
	  // fprintf(outfile, "strengths[%d] = %f\n", i, strengths[i]);
	  // fprintf(outfile, "totals[%d] = %f\n", i, totals[i]);
	  // fprintf(outfile, "maxima[%d] = %f\n", i, maxima[i]);
	  // fprintf(outfile, "strengths[%d] / %d = %f\n", i, pow_int(2, i), strengths[i] / pow(2, i));
	  fflush(outfile);
	}
  }

  fprintf(outfile, "The averages below represent power per DC bin.\n");
  for(int i = 0; i < NUM_STRENGTHS; i++) {
	averages[i] = totals[i] / (float) num_ffts / (float) pow_int(2,i) / (float) fft_len2 / 2.0;
	fprintf(outfile, "averages[%d] = %f\n", i, averages[i]);
  }
  for(int i = 0; i < NUM_STRENGTHS; i++) {
	fprintf(outfile, "normalized maxima[%d] = %f\n", i, maxima[i] / (float) pow_int(2,i) / (float) fft_len2 / 2.0);
  }
  fflush(outfile);
}

int Science::get_indices_to_randomize(std::vector<long>& indices, int scale) {
  FILE *outfile;
  state_t &state=client.ap_shmem->ap_gdata.state;
  int init_indices = 100;
  int num_indices = 0;
  int size = init_indices;
  indices.resize(init_indices);
  if(scale > 7) {
	  fprintf(stderr, "In Science::get_indices_to_randomize: error: scale > 7\n");
	  printf("In Science::get_indices_to_randomize: error: scale > 7\n");
	  exit(-1);
  }
  // power_threshold[scale]
  float power_threshold[] = { 12.1, 6.9, 5.3, 3.0, 2.46, 1.96, 1.62, 1.37};
  int fft_len2 = 1024;
  long datasize = 8 * 1024 * 1024;
  float strength;
  outfile = fopen("indices.txt", "w");

  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
#if (__SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *) i_data,  fft_len2))

/*#elif _GNU_SOURCE && (__SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *) i_data,  fft_len2))*/

#else
	if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, fft_len2))
#endif
	{
	  exit(SBTF_ERR);
	}
#if USE_FFTW
	fftwf_execute(fp2); // Note that this fft knows about fft_len2 (if you define it in astropulse.h), and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024
#endif
#if USE_ACML
	cfft1d(-1,acmlfftlen2,(complex*)data,fp2,&info);
#endif
	/* For debugging: check power in DC bin for this data chunk
	if(data_chunk_now == 1359872) {
		printf("DC bin: %f %f, DC power: %f\n", data[0][0], data[0][1], (data[0][0] * data[0][0] + data[0][1] * data[0][1]) / fft_len2 / 2.0);
	}*/

	int expscale = pow_int(2, scale);
	strength = 0;
	for(int j = - expscale / 2; j < (expscale + 1) / 2; j++) {
	  int j2 = (j + fft_len2) % fft_len2;
#if !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
	  float real = data[j2][0];
	  float imag = data[j2][1];
#elif ( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
	  float real = ((ap_complex*)data)[j2][0];
	  float imag = ((ap_complex*)data)[j2][1];
#endif
	  strength += (real * real + imag * imag) / fft_len2 / 2.0 / expscale;
	  // strength += (real * real + imag * imag) / fft_len2 / expscale;
	}

	/* For debugging: check that total power / 2.0 = fft_len2
	if(data_chunk_now < 10000) {
	float total = 0;
	for(int i = 0; i < fft_len2; i++) {
		total += (data[i][0] * data[i][0] + data[i][1] * data[i][1]) / fft_len2 / 2.0;
	}
	printf("total = %f\n", total);
	}
	*/

	if(strength > power_threshold[scale]) {
	  fprintf(outfile, "At sample %ld, data_chunk_now %ld, strength %f > %f\n", data_chunk_now * 4, data_chunk_now, strength, power_threshold[scale]);
	  /* Given 3 index finds, if the blanking range of the first overlaps the range
	   * of the third, the one in the middle is redundant so overwrite it rather than
	   * extending the list.
	   */
	  if(num_indices > 1 && (indices[num_indices - 2] + 100000) >= (data_chunk_now - 100000))
		num_indices--;
	  indices[num_indices++] = data_chunk_now;
	  if(num_indices >= size) {
		size = num_indices + init_indices;
		indices.resize(size);
	  }
	}
  }
/*  if (num_indices) {
	fprintf(outfile, "\nRetained data_chunk indices:\n");
	for (int k = 0; k < num_indices; k++) fprintf(outfile, "%ld\n", indices[k]);
  }*/
  if (num_indices) {
	long FHalf = state.fft_len/2;
	long RHalf = FHalf/4;
	long HiB, LoB, LoNB;

	fprintf(outfile, "\nRetained data_chunk indices blanking summary:\n");
	fprintf(outfile, "   index    effect\n");
	fprintf(outfile, " -------    ---------------------------------------\n");
	if (indices[0] > 100000)
	  fprintf(outfile, "          unblanked: fft_num %8ld thru %-8ld\n", 0L,((indices[0]-100000)/RHalf)*FHalf);
	for (int k = 0; k < num_indices; k++)  {
	  LoB = std::max(0L, ((((long)indices[k]-100000)/RHalf)+1)*FHalf);
	  HiB = std::min(datasize*4-2*FHalf, (((long)indices[k]+100000)/RHalf)*FHalf);
	  LoNB = datasize*4-FHalf;
	  if (k < num_indices - 1) LoNB = (((long)indices[k+1]-100000)/RHalf)*FHalf;
	  fprintf(outfile, "%8ld    blanked: fft_num %8ld thru %-8ld\n", indices[k], LoB, HiB);
	  if (HiB < LoNB-FHalf)
		fprintf(outfile, "          unblanked: fft_num %8ld thru %-8ld\n", HiB+FHalf, LoNB-FHalf);
	}
  }

  fflush(outfile);
  fclose(outfile);

  //JWS: Early exit if there's insufficient unblanked data. Logic from generate_envelope()
  // moved here for slightly quicker exit if noise envelope method is in use.
  // Also so BLANKIT need not invoke that code, and the short FFA can be disabled if
  // all data for it is blanked.
  int num_ffts_forecast = 0;
  bool good_data_chunk, goodlow = false;
  int data_chunk_size = state.fft_len / 4;
  for(int data_chunk=0; data_chunk <= state.datasize - data_chunk_size; data_chunk+= data_chunk_size) {
	good_data_chunk = true;     // Check whether this is a bad data chunk (within 100000 bytes of RFI.)
	for(int n=0; n<num_indices; n++) {
	  if( indices[n] < data_chunk + 100000 && indices[n] > data_chunk - 100000) {
		good_data_chunk = false;
		break;
	  }
	}
	if(good_data_chunk) {
	  num_ffts_forecast += state.fft_len / fftlen_envelope;
#ifdef BLANKIT
	  if (data_chunk < state.fold_buf_bytes_short) goodlow = true;
#endif
	}
  }
  int min_ffts = 100;
  if(num_ffts_forecast < min_ffts) {
	printf("In ap_remove_radar.cpp : get_indices_to_randomize: num_ffts_forecast < %d.  Blanking too much RFI?\n", min_ffts);
	fprintf(stderr, "In ap_remove_radar.cpp: get_indices_to_randomize: num_ffts_forecast < %d.  Blanking too much RFI?\n", min_ffts);
	// Quit early
	state.numSamplesBlanked = 1;    // forcing fraction_blanked to 1.0 dividend
    state.lastDataChunkBlanked = 1; // divisor
	fflush(stdout);
	client.outfile.write_main(client.signal_vector);
	client.outfile.finish_results();
	client.finish();
	main_timer.stop();
	boinc_finish(0);
	exit(0);
  }
#ifdef BLANKIT
  if (!goodlow) {
	skipffa_short = true;         // don't process short
	state.fold_level_small = 99;  // inhibit short save
  }
#endif
  return num_indices;
}

#ifndef BLANKIT


// Read in as many data chunks as needed from client.rawdata.
// Warning: this function messes with the Science data array.
void Science::generate_envelope(const std::vector<long>& indices, int num_indices) {
  // initialize the envelope array
  envelope = new double[fftlen_envelope];
  for(int i=0; i<fftlen_envelope; i++)
	 envelope[i] = 0;

  // GOAL: find envelope (as it should be after one-bitting)

  // * cycle through the #s in the data, adding each of them to
  //   envelope_part
  //   env_index tells which envelope_part we are on
  bool good_data_chunk;
  int data_chunk_size = client.ap_shmem->ap_gdata.state.fft_len / 4;
  for(int data_chunk=0; (unsigned)data_chunk <= client.rawdata.size() - data_chunk_size; data_chunk+= data_chunk_size) {

	// Check whether this is a bad data chunk (within 100000 bytes of RFI.)
	good_data_chunk = true;
	for(int n=0; n<num_indices; n++) {
	  if( indices[n] < data_chunk + 100000 && indices[n] > data_chunk - 100000) {
		good_data_chunk = false;
		break;
	  }
	}
	if(!good_data_chunk) continue;

	// Extract our data chunk into the data array.
	splitter_bits_to_float(&client.rawdata[data_chunk], (float*)data, client.ap_shmem->ap_gdata.state.fft_len);
	for(int env_index=0; env_index<client.ap_shmem->ap_gdata.state.fft_len / fftlen_envelope; env_index++) {
		for(int i = 0; i < fftlen_envelope; i++) {
		  int d = env_index * fftlen_envelope + i;
		  envelope_part[i][0] = ((ap_complex*)data)[d][0];
		  envelope_part[i][1] = ((ap_complex*)data)[d][1];
		}
#if USE_FFTW
	  fftwf_execute(fp_envelope);
#endif
#if USE_ACML
	  cfft1d(-1,acmlfftlenenv,(complex*)envelope_part,fp_envelope,&info);
#endif
	  for(int i = 0; i < fftlen_envelope; i++) {
		envelope[i] += envelope_part[i][0] * envelope_part[i][0]
					 + envelope_part[i][1] * envelope_part[i][1];
	  }
	}
  }

  // * when you're done, normalize the power array:
  //    * add up the powers, and divide by fftlen_envelope
  //      to get the average power
  //    * divide the whole array by the average power
  //    * array is avp(nu), the average power at frequency nu, normalized
  //      so that <avp(nu)>_nu = 1.
  double total_power = 0;
  for(int i = 0; i < fftlen_envelope; i++)
	total_power += envelope[i];
  double avg_power = total_power / fftlen_envelope;
  for(int i = 0; i < fftlen_envelope; i++)
	envelope[i] /= avg_power;
}

void Science::generate_pre_envelope() {

  double known_alpha = 0.547539;

  double total_power=0; //uje: preventing uninitialized usage
  pre_envelope = new double[fftlen_envelope];

  for(int i = 0; i < fftlen_envelope; i++) {
	pre_envelope[i] = (1 + known_alpha) * envelope[i] - known_alpha;
	if(pre_envelope[i] < 0) pre_envelope[i] = 0;
	total_power += pre_envelope[i];
  }
  if(total_power <= 0) {
	 printf("Error in ap_remove_radar.cpp: generate_pre_envelope: total <= 0.\n");
	 fprintf(stderr, "Error in ap_remove_radar.cpp: generate_pre_envelope: total <= 0.\n");
	 exit(1);
  }
  for(int i = 0; i < fftlen_envelope; i++) // set average power per bin to 1.
	pre_envelope[i] *= (fftlen_envelope / total_power);

  //JWS: Modified code moved from generate_time_noise_with_envelope()
  // generates the patch multipliers one time rather than millions
  //  * multiply patch by sqrt(avp(nu)).
  //      expected power in bin nu is now avp(nu)
  for(int k = 0; k<fftlen_patch; k++) {
	// 0.5 is added so we get 0.5, ... fftlen_noise - 0.5, which is then rounded down.
	int index = (int) ( ((float)k + 0.5) / (float)fftlen_patch * (float)fftlen_envelope);
	patch_mults[k] = sqrt(pre_envelope[ index ]);

	// if(k == 0)
	//  printf("noise[0][0] = %f, noise[0][1] = %f, noise power = %f, multiplier = %f, index = %d\n", noise[0][0], noise[0][1], noise[0][1] * noise[0][1] + noise[0][0] * noise[0][0], multiplier, index);
  }
  //JWS: Delete arrays no longer needed.
  delete[] envelope;
  delete[] pre_envelope;
}

void Science::randomize_indices(const std::vector<long>& indices, int num_indices, int data_chunk_now, int fft_len, state_t& state) {

// FILE* randsampfile = fopen("randsamp.txt", "a");
  int dcn_up  = data_chunk_now + 100000;
  int dcn_dwn = data_chunk_now - 100000;

  /* For debugging: print values of bits before they are randomized.
  if(data_chunk_now == 1359872 - 4096) {
	  printf("data_chunk_now: 1359872 - 4096\n");
	  for(int k = 16384; k < 16384 + 1024; k++) {
		  printf("%f %f | ", data[k][0], data[k][1]);
		  if(k % 16 == 0) printf("\n");
	  }
  }
  */

  // printf("data_chunk_now = %d\n", data_chunk_now); // for debugging

  // The count of data chunks blanked divided by all data chunks is blanking ratio.
  state.lastDataChunkBlankedNew++ ; //increment all count

  for(int current = 0; current < num_indices; current++) {
	  // If we're within 100000 bytes of a DC pulse, randomize this data chunk.
	  if(indices[current] < dcn_up && indices[current] > dcn_dwn) {
	  int debug = 0;
	  state.numSamplesBlankedNew++ ; //increment blanked count
	  randomize_samples(fft_len, state.mtr_new, debug);
	  // fprintf(randsampfile, "Num chunks blanked: %d, Total chunks: %d, Percent blanked: %f\n",
	  // state.numSamplesBlankedNew, state.lastDataChunkBlankedNew,
	  // 100.0f * (float)state.numSamplesBlankedNew / (float)state.lastDataChunkBlankedNew);
	  break;
	}
  }
  // fclose(randsampfile);
}



double gasdev(MtRand &mtr) {
  static int iset=0;
  static double gset;
  double fac, rsq, v1, v2;
  if(iset == 0) {
	do {
//	  v1=2.0*mtr.drand32()-1.0;
//	  v2=2.0*mtr.drand32()-1.0;
//JWS: retrieve as 32 bits of unsigned long. Toggling the 2^31 bit and casting to
// signed integer gives -2^31 to +2^31-1 range, replacing the -1.0 offsetting
// operation. The xor is faster, so is converting from signed integer to double.
	  v1=((int)(mtr.random32()^0x80000000))*(1.0/2147483648.0);
	  v2=((int)(mtr.random32()^0x80000000))*(1.0/2147483648.0);
	  rsq=v1*v1+v2*v2;
	} while(rsq >= 1.0 || rsq==0.0);
	fac=(1/rsq)*sqrt(-2.0*log(rsq)*rsq); //uje: costly, possible cause of Ubuntu904 failure ?
	gset=v1*fac;
	iset=1;
	return v2*fac;
  } else {
	iset=0;
	return gset;
  }
}

// One-bits the data (called "noise") which has length fftlen_noise
// That is, values greater than 0 are set to 1, and less than 0 are
// set to -1.
void one_bit(fftwf_complex * noise, int fftlen_noise) {
  int sum_0_0 = 0;
  int sum_0_1 = 0;
  for(int k=0; k<fftlen_noise; k++) {
	noise[k][0] = (noise[k][0] < 0) ? -1 : 1;
	noise[k][1] = (noise[k][1] < 0) ? -1 : 1;
	sum_0_0 += noise[k][0];
	sum_0_1 += noise[k][1];
  }
  // printf("noise[0][0] = %f, noise[0][1] = %f\n", noise[0][0], noise[0][1]);
  // printf("sum_0_0 = %d, sum_0_1 = %d\n", sum_0_0, sum_0_1);
}

// This generates unnormalized time domain noise with frequency spectrum given by "envelope".
// (Because it will be one-bitted immediately afterwards.)
#if USE_FFTW
void generate_time_noise_with_envelope(fftwf_complex * noise, double * patch_mults, int fftlen_noise, int fftlen_envelope, fftwf_plan fp_noise_inv, MtRand &mtr) {
  // double rnd1;
  // double rnd2;
  // double HALF_RAND_MAX = RAND_MAX/2.0;
  // double multiplier;

  //  * generate fftlen_patch random cplx bit pairs (as doubles)
  /*
  for(int k=0; k<fftlen_noise; k++) {
	rnd1=static_cast<double>(rand());
	rnd1=rnd1 <= HALF_RAND_MAX ? -1 : 1;
	rnd2=static_cast<double>(rand());
	rnd2=rnd2 <= HALF_RAND_MAX ? -1 : 1;
	noise[k][0]=rnd1;
	noise[k][1]=rnd2;
  }*/
  //  * fourier transform them to get a Gaussian distribution
  //      average power per frequency bin is some constant P, and
  //      expected power in each bin is <P>.
  // fftwf_execute(fp_noise);
  // avg power per bin = 2 * fftlen_patch.
  //  * TODO: divide by sqrt(P), so avg power in each bin is 1.
  /* double sqrt_avg_power = sqrt((double) 2 * fftlen_noise);
  for(int k = 0; k<fftlen_noise; k++) {
	noise[k][0] /= sqrt_avg_power;
	noise[k][1] /= sqrt_avg_power;
  } */
//JWS: Combine into next loop
/*  for(int k=0; k<fftlen_noise; k++) {
	noise[k][0] = gasdev(mtr);
	noise[k][1] = gasdev(mtr);
  }
*/
  //  * multiply patch by sqrt(avp(nu)).
  //      expected power in bin nu is now avp(nu)
  for(int k = 0; k<fftlen_noise; k++) {
/*JWS: moved rounded index and sqrt() to generate_pre_envelope()
	// 0.5 is added so we get 0.5, ... fftlen_noise - 0.5, which is then rounded down.
	int index = (int) ( ((float)k + 0.5) / (float)fftlen_noise * (float)fftlen_envelope);
	multiplier = sqrt(envelope[ index ]);

	noise[k][0] *= multiplier;
	noise[k][1] *= multiplier;
*/
	noise[k][0] = gasdev(mtr) * patch_mults[k];
	noise[k][1] = gasdev(mtr) * patch_mults[k];
	// if(k == 0)
	//  printf("noise[0][0] = %f, noise[0][1] = %f, noise power = %f, multiplier = %f, index = %d\n", noise[0][0], noise[0][1], noise[0][1] * noise[0][1] + noise[0][0] * noise[0][0], multiplier, index);
  }
  //  * inverse FFT
  fftwf_execute(fp_noise_inv);
}
#endif
#if USE_ACML
void generate_time_noise_with_envelope(ap_complex * noise, double * envelope, int fftlen_noise, int fftlen_envelope, complex * fp_noise_inv, MtRand &mtr) {

  // double rnd1;
  // double rnd2;
  // double HALF_RAND_MAX = RAND_MAX/2.0;
  double multiplier;
  int info;

  //  * generate fftlen_patch random cplx bit pairs (as doubles)
  /*
  for(int k=0; k<fftlen_noise; k++) {
	rnd1=static_cast<double>(rand());
	rnd1=rnd1 <= HALF_RAND_MAX ? -1 : 1;
	rnd2=static_cast<double>(rand());
	rnd2=rnd2 <= HALF_RAND_MAX ? -1 : 1;
	noise[k][0]=rnd1;
	noise[k][1]=rnd2;
  }*/
  //  * fourier transform them to get a Gaussian distribution
  //      average power per frequency bin is some constant P, and
  //      expected power in each bin is <P>.
  // fftwf_execute(fp_noise);
  // avg power per bin = 2 * fftlen_patch.
  //  * TODO: divide by sqrt(P), so avg power in each bin is 1.
  /* double sqrt_avg_power = sqrt((double) 2 * fftlen_noise);
  for(int k = 0; k<fftlen_noise; k++) {
	noise[k][0] /= sqrt_avg_power;
	noise[k][1] /= sqrt_avg_power;
  } */
  for(int k=0; k<fftlen_noise; k++) {
	noise[k][0] = gasdev(mtr);
	noise[k][1] = gasdev(mtr);
  }
  //  * multiply patch by sqrt(avp(nu)).
  //      expected power in bin nu is now avp(nu)
  for(int k = 0; k<fftlen_noise; k++) {
	// 0.5 is added so we get 0.5, ... fftlen_noise - 0.5, which is then rounded down.
	int index = (int) ( ((float)k + 0.5) / (float)fftlen_noise * (float)fftlen_envelope);
	multiplier = sqrt(envelope[ index ]);

	noise[k][0] *= multiplier;
	noise[k][1] *= multiplier;
	// if(k == 0)
	//  printf("noise[0][0] = %f, noise[0][1] = %f, noise power = %f, multiplier = %f, index = %d\n", noise[0][0], noise[0][1], noise[0][1] * noise[0][1] + noise[0][0] * noise[0][0], multiplier, index);
  }
  //  * inverse FFT
  //fftwf_execute(fp_noise_inv);
  cfft1d(1,fftlen_noise,(complex *)noise,fp_noise_inv,&info);
}
#endif

// Use debug flag if you want randomize_samples to print something out
// on specific samples, for instance.
void Science::randomize_samples(int fft_len, MtRand& mtr, int debug) {
  // FILE *randsampfile;
  // randsampfile = fopen("randsamp.txt", "a");
  // fprintf(randsampfile, "in randomize_samples: in: %s\n", mtr.toString());
  int patch_index = 0;

  for(int index = 0; index < fft_len; index++) {
	if(patch_index == 0) {
	  generate_time_noise_with_envelope(patch, patch_mults, fftlen_patch, fftlen_envelope, fp_patch_inv, mtr);
	  one_bit(patch, fftlen_patch);
	}
// * use bits from this distribution to fill the holes
	while(index < client.ap_shmem->ap_gdata.state.fft_len) {
#if USE_FFTW // was _WIN32 && USE_VIRTUAL_ALLOC
	  ((ap_complex*)temp_in)[index][0] = ((ap_complex*)patch)[patch_index][0];
	  ((ap_complex*)temp_in)[index][1] = ((ap_complex*)patch)[patch_index][1];
#else
	  ((ap_complex*)data)[index][0] = ((ap_complex*)patch)[patch_index][0];
	  ((ap_complex*)data)[index][1] = ((ap_complex*)patch)[patch_index][1];
#endif
	// data[index][0] = mtr.drand32() > 0.5 ? 1.0 : - 1.0;
	// data[index][1] = mtr.drand32() > 0.5 ? 1.0 : - 1.0;
	  patch_index++;
	  if(patch_index >= fftlen_patch) {
// * when you don't have any patch bits left,
// start a new patch, going back through the for loop
		patch_index = 0;
		break;
	  }
	  index++;
	}
	patch_index = 0;
  } // end "for(int index = 0; index < fft_len; index++)"
  // fprintf(randsampfile, "in randomize_samples: out: %s\n", mtr.toString());
  // fflush(randsampfile);
  // fclose(randsampfile);
}
#endif // BLANKIT

} // end namespace Astropulse
