// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef H_ASTROPULSE
#define H_ASTROPULSE
#define DEBUGGING_0 // Tells the client to regularly inform stdout about its progress
#define POWER_MAX 1000
#define DM_MAX 16000

// #define TOTAL_FLOPS 6.21524e+14
// #define TOTAL_FLOPS 1.1187432e+15
// Need to adjust for variable numbers of DMs
#define FLOPS_PER_DM 7.93210e+10

#include "ctype.h"
#include "stdio.h"
#include <string>

#if _GNU_SOURCE || __MINGW32__ || __APPLE__ || __FreeBSD__
#include <stdint.h> //uje: for __rdtsc replacement
#include <typeinfo> //uje: for using "typeid" with gcc
#include <fstream> //uje: for using "ifstream" with gcc
#endif

#include "intrinsics.h"
#if USE_GPU
	#include "cuda_runtime_api.h"
	#include "cufft.h"
#endif
#if USE_FFTW
	#include "fftw3.h"
#endif
#if USE_IPP
	#include "ipp_static.h"
#endif
#if USE_ACML
	#include "acml.h"
#endif
#include "boinc_api.h"
#include "filesys.h"
#include "mfile.h"
#include "db_table.h"
#include "ap_schema.h"

#include "ap_const.h"

#include "ap_time_series.h"

#if (_WIN32 && _MSC_VER)
#pragma intrinsic(__rdtsc)
extern "C" {
	unsigned __int64 __rdtsc(void);
}
#elif __MINGW32__ || (_GNU_SOURCE && (__WORDSIZE == 32))
inline uint64_t __rdtsc( void) {
    uint64_t cycles;
		// Use CPUID to stop the instruction pipeline and get starting cycle
    __asm volatile ("cpuid \n rdtsc" : "=A" (cycles) : "a" (0) : "%ebx", "%ecx" );
    return (cycles);
}
#elif (_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)
inline uint64_t __rdtsc() {
	uint32_t lo, hi;
	asm volatile ("xorl %%eax,%%eax \n        cpuid"
		::: "%rax", "%rbx", "%rcx", "%rdx");
	asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));
	return (uint64_t)hi << 32 | lo;
}
#endif // _WIN32 && _MSC_VER

#if _MSC_VER
#if _WIN64
static inline int lrint (double const x) {
	return _mm_cvtsd_si32(_mm_load_sd(&x));}
#else
static inline int lrint (double const x) {
	int n;
	__asm fld qword ptr x;
	__asm fistp dword ptr n;
	return n;}
#endif // _WIN64
#endif // _MSC_VER

extern bool debug_msg;
namespace Astropulse {

#ifndef ENUM_CODE_SEGMENT_NAME
#define ENUM_CODE_SEGMENT_NAME
enum code_segment_name {main_ffa_long, main_ffa_short, build_threshold_table_long, build_threshold_table_short, single_pulse};
#endif

};

#include "ap_state.h"
#include "gdata.h"
#include "time.h"

namespace Astropulse {
using std::string;

enum OnOff {
  on, off
};

const double frac_in_main_ffa[10] = {0.895, 0.89, 0.87, 0.85, 0.8,
                                   // 0      1     2     3     4
                                     0.7, 0.6, 0.4, 0.2, 0.05};
                                   // 5    6    7    8    9

const double frac_in_btt[10] = {0.1,   0.1,  0.1,  0.1,  0.1,
				   // 0      1     2     3     4
				   0.1, 0.1, 0.1, 0.1, 0.05};
				   // 5    6    7    8    9

const double frac_in_main_ffa_short = 0.2;
//const unsigned fft_len1=32768;//R: if current WU parameters will be changed this value should be corrected or code should be reverted to variable FFT length.

inline int int_log2(int input) {
	int i = 0;
	while(input >>= 1) i++;
	return i;
}
/*
char* to_c_string(const string& str) {
	char* temp= new char[str.length()+1];
	str.copy(temp, string::npos);
	temp[str.length()]=0;
	return temp;
}*/

class timer {
	private:
	clock_t ticks;
	clock_t last_clock;
	clock_t start_clock;

	long seconds;
	long last_time;
	long start_seconds;

	OnOff state;

	public:
	timer();
	void start();
	void stop();
	void reset();
	clock_t get_ticks(){return ticks;};
	clock_t get_total_ticks();
	long get_seconds();
	long get_total_seconds();
	void print_total_time();
};

template<class T,class TT> class Counters{
	static TT total;
	static TT min;
	static TT max;
#if _MSC_VER
	static unsigned __int64 count;
#elif ((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__
	static uint64_t  count;
#else
	static unsigned long  count;
#endif
public:
	static void print();
	static void update(TT value);
};

template <typename T,typename TT>
TT  Counters<T,TT>::total=0;
template <typename T,typename TT>
TT  Counters<T,TT>::min=0;
template <typename T,typename TT>
TT  Counters<T,TT>::max=0;
template <typename T,typename TT>
#if _MSC_VER
unsigned __int64 Counters<T,TT>::count=0;
#elif ((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__
uint64_t  Counters<T,TT>::count=0;
#else
unsigned long Counters<T,TT>::count=0;
#endif
template <typename T,typename TT>
inline void Counters<T,TT>::update(TT value){
		total+=value;
		if(count){
		min=(min>value)?value:min;
		max=(max<value)?value:max;}
		else{min=max=value;}
		count++;
}
template <typename T,typename TT>
#if _MSC_VER
inline void Counters<T,TT>::print(){
	fprintf(stderr,"%s:\ttotal=%.4g,\tN=%I64u,\t<>=%.4g,\tmin=%.4g\tmax=%.4g\n",
		typeid(T).name(),(float)total,(float)count,(float)((count)?(total/count):total),(float)min,(float)max);
}
#elif ((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__
inline void Counters<T,TT>::print(){
	char* tidn = (char*)typeid(T).name();
	string cl = " "; //uje: "class "
	if (tidn[0]=='N' && tidn[1]=='1' && tidn[2]=='0') {//uje: linux specific JWS: GCC specific
		string temp = "";
		for(int i=3; tidn[i]!='\0'; i++)  temp += tidn[i];
		cl.append(temp);
	}
	size_t i2 = cl.find("T_"); //uje: i1 = cl.find_first_not_of("class Astropulse"),
 	cl.erase(0,i2+2);
	cl.replace(cl.rfind("E"),1,"");
	fprintf(stderr,"%-19s total=%-10.4E, N=%-12lu, <>=%-10.4G, min=%-10.4G, max=%-10.4G\n",
		cl.c_str(),(double)total,(unsigned long)count,(double)((unsigned long)(count)?(double)(total/count):(double)total),(double)min,(double)max);
}
#elif _GNU_SOURCE && (__WORDSIZE == 32)
//missing
#else
inline void Counters<T,TT>::print(){
	fprintf(stderr,"%s: total=%.4g, N=%-12u, <>=%.4g, min=%.4g, max=%.4g\n",
		typeid(T).name(),total,count,((count)?(total/count):total),min,max);
}
#endif
//class T_AveragePower{}; //no longer in use
class T_per_int_short{};
class T_per_int_long{};
class T_n_per_short{};
class T_n_per_long{};
class T_FFA_single_pulse{};
class T_FFA_multiple_pulse{};

template<class T> class Timings{
#if _MSC_VER
	static unsigned __int64  total;
	static unsigned __int64  count;
	static unsigned __int64  min_delta;
	static unsigned __int64  max_delta;
	unsigned __int64 start;
#elif ((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__
	static uint64_t  total;
	static uint64_t  count;
	static uint64_t  min_delta;
	static uint64_t  max_delta;
	uint64_t start;
#elif _GNU_SOURCE && (__WORDSIZE == 32)
//missing
#else
	static unsigned long  total;
	static unsigned long  count;
	static unsigned long  min_delta;
	static unsigned long  max_delta;
	unsigned long  start;
#endif
public:
	Timings();
	~Timings();
	static void print();
};
#if _MSC_VER
template <typename T>
unsigned __int64 Timings<T>::total=0;
template <typename T>
unsigned __int64  Timings<T>::min_delta=(unsigned __int64)-1;
template <typename T>
unsigned __int64  Timings<T>::max_delta=0;
template <typename T>
unsigned __int64  Timings<T>::count=0;
#elif ((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__
template <typename T>
uint64_t  Timings<T>::total=0;
template <typename T>
uint64_t  Timings<T>::min_delta=(uint64_t)-1;
template <typename T>
uint64_t  Timings<T>::max_delta=0;
template <typename T>
uint64_t  Timings<T>::count=0;
#elif _GNU_SOURCE && (__WORDSIZE ==32)
//missing
#else
template <typename T>
unsigned long  Timings<T>::total=0;
template <typename T>
unsigned long  Timings<T>::min_delta=(unsigned long)-1;
template <typename T>
unsigned long  Timings<T>::max_delta=0;
template <typename T>
unsigned long  Timings<T>::count=0;
#endif
template <typename T>
inline Timings<T>::Timings(){
#if _WIN32
#if __SSE__ || __MINGW32__
	start=__rdtsc();
#else
	start=GetTickCount();
#endif
#elif _GNU_SOURCE || __APPLE__ || __FreeBSD__
#if __SSE__ && (__WORDSIZE == 64)
	start=__rdtsc();
#elif __SSE__ && (__WORDSIZE == 32)
      	//missing
#else
	start=GetTickCount();
#endif
#endif
}
template <typename T>
inline Timings<T>::~Timings(){
#if __MINGW32__
	uint64_t delta=__rdtsc()-start;
#elif _WIN32
#if __SSE__
	register unsigned __int64  delta=__rdtsc()-start;
#else
	register unsigned __int64  delta=GetTickCount()-start;
#endif
#elif _GNU_SOURCE || __APPLE__ || __FreeBSD__
#if __SSE__ && (__WORDSIZE == 64)
	register uint64_t  delta=__rdtsc()-start;
#elif __SSE__ && (__WORDSIZE == 32)
	//missing
#else
	register uint64_t  delta=GetTickCount()-start;
#endif
//#else	register unsigned long  delta=__rdtsc()-start;
#else
	register unsigned long  delta=GetTickCount()-start;
#endif
#if !(_GNU_SOURCE && (__WORDSIZE == 32))
		total+=(delta);
		min_delta=(min_delta>delta)?delta:min_delta;
		max_delta=(max_delta<delta)?delta:max_delta;
		count++;
#endif
}

template <typename T>
inline void Timings<T>::print(){
#if __SSE__ && !_GNU_SOURCE && !__APPLE__ && !__FreeBSD__ && !__MINGW32__
	if (count > 1) {
	  fprintf(stderr,"%s:\ttotal=%.2e,\tN=%I64u,\t<>=%I64u (%.2e),\tmin=%I64u (%.2e),max=%.2e\n",
		typeid(T).name(),(float)total,count,total/count,(float)((count)?(total/count):total),min_delta,(float)min_delta,(float)max_delta);
	} else {
	  fprintf(stderr,"%s:\ttotal=%.2e,\tN=%I64u\n", typeid(T).name(),(float)total,count);
	}
#elif __SSE__ && (((_GNU_SOURCE || __APPLE__ || __FreeBSD__) && (__WORDSIZE == 64)) || __MINGW32__)
	char* tidn = (char*)typeid(T).name();
	string cl = " "; //uje: "class "
	if (tidn[0]=='N' && tidn[1]=='1' && tidn[2]=='0') {//uje: linux specific
		string temp="";
		for(int i=3; tidn[i]!='\0'; i++)  temp += tidn[i];
		cl.append(temp);
	}
	size_t i2 = cl.find("T_"); //uje:  i1 = cl.find_first_not_of("class Astropulse"),
	cl.erase(0,i2+2);
	cl.replace(cl.rfind("E"),1,"");
	if (count > 1) {
	  fprintf(stderr,"%-19s total=%-10.3E, N=%-12lu, <>=%-10.3E, min=%-10.3E, max=%-10.3E\n", cl.c_str(),
		  (double)total,(unsigned long)count,(double)(total/count),(double)min_delta,(double)max_delta);
	} else {
	  fprintf(stderr,"%-19s total=%-10.3E, N=%-12lu\n", cl.c_str(), (double)total, (unsigned long)count);
	}
#elif __SSE__ && _GNU_SOURCE && (__WORDSIZE == 32)
//missing
#else
	fprintf(stderr,"%s  total elapsed time=%I64u(ms),  count=%I64u,  mean delta=%I64u(ms),  min delta=%I64u(ms),  max delta=%I64u(ms)\n",
		typeid(T).name(),total,count,(count)?(total/count):total,min_delta,max_delta);
#endif
}

class T_AppInit{};
class T_AppCrunch{};
class T_dechirp{};
class T_ffa{};

//R:these counters for using inside FFA function, mostly for GPU builds, but can be used in CPU FFA too
class T_FFA_fetch{};
class T_FFA_tt_build{};
class T_FFA_stride_add{};
class T_FFA_coadd{};
class T_FFA_coadd_inner{};
class T_FFA_compare{};
class T_FFA_compare_inner{};
template<int size> class T_FFA_cc_{};
class T_FFA_signal_merge{};
class T_Stream{};//all stream operations like creating, destroying, copying and reading

class T_coadd{};
class T_FFT_forward{};
class T_FFT_inverse{};
//class T_filter{}; //no longer used

#if USE_CUDA
class T_GPU_data_put{};
class T_GPU_data_get{};
#endif
extern void* malloc_a(size_t size);
extern void free_a(void* ptr);


struct time_series_b_t {
	enum { MAXLENGTH = 64 };
	std::vector<unsigned char> data;
	int length;

	time_series_b_t() : length(0) {
		data.reserve(MAXLENGTH);
	}
	void build(float array[], int index, int range, float n_bins);
};

enum ERROR_CODES {
	FWRITE_ERR = -1,	// Error writing a file.
	SBTF_ERR   = -2,	// Error from splitter_bits_to_float
	FOPEN_ERR  = -3,
	PARSE_ERR  = -4,
	FREAD_ERR  = -5
};

class Wufile {
	std::ifstream wufile;
	public:
	void init();
	void parse_header_to_state();
	void parse_header_for_writeout();
	void read_raw_data();
	void finish();
};

void debug_loop(int level, int number, bool debug_loop_msg);
void read_signal_vector(std::vector<ap_signal> &signal_vector);
void read_signal_vector(std::vector<ap_signal> &signal_vector,int use_set);
bool add_signal(std::vector<ap_signal> &signal_vector, ap_signal s);
bool is_blank(ap_signal s);
bool similar_ap_signals(ap_signal s1, ap_signal s2);
int pow_int(int a, int b);

class Outfile {
	/* MFILE is a boinc API construct that can be used to
	* write to output and state files.  It "facilitates atomic
	* checkpoint" meaning it allows output to be written as
	* a single operation (instead of separate character writes?)
	*/
	MFILE outfile;
  public:
	bool pulse_array_dirty;
	bool old_pulse_array_dirty;
	Outfile():pulse_array_dirty(false),old_pulse_array_dirty(false){};
	void init();
	void init(int use_set);
	void finish_results();
	void finish();

	void output_result(const ap_signal &s);
	void write_header();
	void write_main(std::vector<ap_signal> signal_vector);
	void write_main(std::vector<ap_signal> signal_vector,int use_set);
	void flush() {
		outfile.flush();
	}
};

class Statefile {
	FILE *f1,*f2;
	FILE* statefile;
	bool current_set;
  public:
	bool resumed;
	Statefile();
	~Statefile();
	void Read();
	void write();
	void finish();
	void SwitchSet(){
		if(statefile==f1)statefile=f2;
		else statefile=f1;
		current_set=!current_set;
	}
	int CurrentSet(){
		if(current_set)return 1;
		else return 0;
	}

};

class Foldfile {
	FILE *f1,*f2;//R: two files for two checkpoint sets
	FILE* foldfile;
  public:
	bool fold_buf_dirty;//R:Attempt to reduce HDD load and save CPU time for more useful things.
	bool old_fold_buf_dirty;//R:Attempt to reduce HDD load and save CPU time for more useful things.
    Foldfile();
	~Foldfile();
	int Read(int cur_set);
	void write();
	void UseSet(int use_set){
		if(use_set==0)foldfile=f1;
		else foldfile=f2;
	}
	void finish();
};

typedef float ap_complex[2];
#define WISDOM_FNAME "wisdom.dat"

class Science {
#if !(_WIN32 && USE_VIRTUAL_ALLOC)
    //R:main app data buffer that will be aligned for maximum memory performance
	float* app_data_array;
#endif
#if (__SSE__ && SPLIT_COMPLEX) //&& !_GNU_SOURCE
	float* data;
	float* i_data;
	float* i_temp;
	float* i_chirp;
	float* temp;
	float* chirp;

/*#elif _GNU_SOURCE && (__SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT COMPLEX work
	float* data;
	float* i_data;
	float* temp;
	float* i_temp;
	float* chirp;
	float* i_chirp;*/

#else
	ap_complex* data;
	ap_complex* temp;
	ap_complex* temp_in_allo;
	ap_complex* temp_in;
	ap_complex* temp_in_neg_allo;
	ap_complex* temp_in_neg;
	int negredy;
	ap_complex* chirp;
#endif
	float* filter_array;
#if USE_GPU
	cufftHandle cuFFTplan;
	cufftComplex* gpudata;
	cufftComplex* gpuchirp;
	cufftComplex* gputemp;
	float* gpupower;
	float2* gpu_filter_array;
	float* gpu_fold_buf_short;
	float* gpu_temp_buf_short;
	float* gpu_temp_coadd_short;
	float* gpu_temp_coadd_long;
	float* gpu_fold_buf_long;
	float* gpu_temp_buf_long;
	cudaError_t gpuError;
	cufftResult_t gpufftError;
#endif
#if USE_FFTW
	fftwf_plan fp, fp2, ip, ipn;
	// fftwf_plan* cp;
	// for remove radar
	int fftlen_patch, fftlen_envelope;
	double * envelope;
	double * pre_envelope;
	double * patch_mults;
	ap_complex* patch;
	ap_complex* envelope_part;
	fftwf_plan fp_patch_inv, fp_envelope;

	class Wisdom {
		char wname[256+10];
	public:
		void load();
		void save();
	};

	Wisdom wisdom;

#endif
#if USE_IPP
	static int order;
	IppsFFTSpec_C_32fc* pFFTSpec;
#endif
#if USE_ACML
	complex* fp;
	complex* fp2;
	complex* ip;
	// fftwf_plan* cp;
	// for remove radar
	int fftlen_patch, fftlen_envelope;
	double * envelope;
	double * pre_envelope;
	double * patch_mults;
	ap_complex* patch;
	ap_complex* envelope_part;
	complex* fp_patch_inv;
	complex* fp_envelope;
	int info;
#endif
  public:
#if USE_ACML
	int acmlfftlen;
	int acmlfftlen2;
	int acmlfftlenpatch;
	int acmlfftlenenv;

#endif
	float* fold_buf_long_pos;
	float* fold_buf_long_neg;
	float* fold_buf_short_pos;
	float* fold_buf_short_neg;
	float* temp_long;
	float* temp_coadd_long;
	float* temp_o2;
	float* temp_o3;
	float* temp_short;
	float* temp_coadd_short;
	unsigned temp_buf_long_size;
	unsigned temp_buf_short_size;
	float tlbt1; // for TLB priming
	float tlbt2;

	void init();
	void mainloop();
	void finish();
  protected:
	void build_chirp_table(int dm_start);
	void convert_bits_to_float();
	void compute_forward_fft();
	void high_pass_filter();
	void dechirp(int dm, float *power, int sign);
	void coadd_in_place(float *power, int size);
	int ffa(float *data, int num_sub_buffers, bool is_long, bool write_lcgf, bool debug_msg, int pulse_limit_single, int pulse_limit_rep, double ffa_thresh_mult);
	void build_threshold_table(int*& table, bool debug_msg, bool is_long, double ffa_thresh_mult);
	int lcgf_binary_search(int min, int max, int ntc, float thresh_lcgf);
	void print_DC_strengths();
	int get_indices_to_randomize(std::vector<long>& indices, int scale);
	void randomize_indices(const std::vector<long>& indices, int num_indices, int data_chunk_now, int fft_len, state_t& state);
	void randomize_samples(int fft_len, MtRand& mtr, int debug);
	void generate_envelope(const std::vector<long>& indices, int num_indices);
	void generate_pre_envelope();
#ifdef PARALLEL
	void dm_chunk_proc();
#endif
}; //class Science

#if USE_FFTW
void generate_time_noise_with_envelope(fftwf_complex * noise, double * envelope, int fftlen_noise, int fftlen_envelope, fftwf_plan fp_noise_inv, MtRand &mtr);
void one_bit(fftwf_complex * noise, int fftlen_noise);
#endif
#if USE_ACML
void generate_time_noise_with_envelope(ap_complex * noise, double * envelope, int fftlen_noise, int fftlen_envelope, complex* fp_noise_inv, MtRand &mtr);
void one_bit(ap_complex * noise, int fftlen_noise);
#endif
double gasdev(MtRand &mtr);

class Client {
  public:
	workunit_header wuheader;
	const char* status;         // Spot to put program status for display.
	// NOTE: any data assigned to `status' must be static data -- if this
	// constraint is not acceptable, change it back to `string' instead
	// of `const char*'.

	Outfile   outfile;
	Wufile    wufile;
	Foldfile  foldfile;
	Statefile statefile;
	AP_SHMEM *ap_shmem;

	Science science;

	std::vector<ap_signal> signal_vector;

	std::vector<unsigned char> rawdata;
	float* power;
	float* fold_buf_long_pos;
	float* fold_buf_short_pos;
	float* fold_buf_long_neg;
	float* fold_buf_short_neg;

	void init();
	void finish();

	inline void do_checkpoint() {
	// outfile.flush();
	// FILE *randsampfile = fopen("randsamp.txt", "a");
	// fprintf(randsampfile, "doing checkpoint\n");
	// fclose(randsampfile);
		if(outfile.pulse_array_dirty)	{
			outfile.write_main(signal_vector,statefile.CurrentSet());
			outfile.pulse_array_dirty=false;
			outfile.old_pulse_array_dirty=true;
		}else	{
			if(outfile.old_pulse_array_dirty){
				outfile.write_main(signal_vector,statefile.CurrentSet());
				outfile.old_pulse_array_dirty=false;
			}else	{//R: can safely ignore pulse file write
				if(debug_msg) fprintf(stderr,"WARNING: Attempt to save unmodified pulse array\n");
			}
		}
		if(foldfile.fold_buf_dirty){
			foldfile.write();
			foldfile.fold_buf_dirty=false;
			foldfile.old_fold_buf_dirty=true;//R:need to rewrite fold file from another checkpoint set.
		}else	{
			//R:we come to checkpoint w/o changes in fold buffers from last checkpoint, but were changes from prev one or not?
			if(foldfile.old_fold_buf_dirty){//R: yes, they were so we need to update fold file anyway
				foldfile.write();
				foldfile.old_fold_buf_dirty=false;//R:now we have both fold files in current state
			}else	{//R: can safely ignore fold file write
				if(debug_msg)  fprintf(stderr,"WARNING: Attempt to save unmodified fold buf\n");
			}
		}
		statefile.write();
		//R:now switching on alternative checkpoint set
		statefile.SwitchSet();
		foldfile.UseSet(statefile.CurrentSet());
	};
	inline void checkpoint() {
		if (boinc_time_to_checkpoint()) {
			do_checkpoint();
			boinc_checkpoint_completed();
		}
	};
};

extern Client client;
extern timer main_timer;
};

#ifndef INFINITY
#define INFINITY (-log(0.0))
#endif

extern bool skipffa_long;
extern bool skipffa_short;
extern double fpopsMult;

#endif
