#include "intrinsics.h"
#if !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
#include "cpu_check.h"
#endif
#if USE_IPP
	#include "ipp_static.h"
#endif
#if _GNU_SOURCE || __APPLE__ || __FreeBSD__
  #include <sys/utsname.h>
  #include <cerrno>
  #include <cstdlib>
  #include <string>
  #include <fstream>
#endif

#ifdef __MINGW32__
static void __cpuid(int buffer[4], int cmd)
	{
    __asm ("cpuid" \
	     : "=a" (buffer[0]), "=b" (buffer[1]), "=c" (buffer[2]), "=d" (buffer[3]) \
	     : "a" (cmd) );
	}
#endif

void intrinsic_cpuid()
{
//R: will print not only host capabilities but app capabilities too.
	fprintf(stderr,"\nBuild features: "); //uje: deleted following ugly tabifications
#ifndef BOINC_APP_GRAPHICS
	   fprintf(stderr,"Non-graphics ");
#endif
#if USE_GPU
	   fprintf(stderr,"CUDA ");
#endif
#if USE_CUDA
	   fprintf(stderr,"CUDA ");
#endif
#if USE_VLAR_KILL
		fprintf(stderr,"VLAR autokill enabled ");
#endif
#if USE_CPU_GPU_TEAM
	   fprintf(stderr,"CPU-GPU team mod ");
#endif
#if USE_CPU_GPU_RESCHEDULING
	   fprintf(stderr,"CPU->GPU re-scheduling enabled ");
#endif
#if USE_AFFINITY_LOCK
		fprintf(stderr,"Affinity lock ");
#endif
#if USE_CROSS_WATCH
		fprintf(stderr,"USE_CROSS_WATCH ");
#endif
#if USE_FFTW
	   fprintf(stderr,"FFTW ");
#endif
#if USE_ACML
	   fprintf(stderr,"ACML ");
#endif
#if USE_IPP
	   fprintf(stderr,"IPP ");
#endif
#if USE_CONVERSION_OPT
		fprintf(stderr,"USE_CONVERSION_OPT ");
#endif
#if USE_INCREASED_PRECISION
	   fprintf(stderr,"USE_INCREASED_PRECISION ");
#endif
#if USE_AMD
	   fprintf(stderr,"AMD specific ");
#endif
#if !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
#if USE_SSE
	   fprintf(stderr,"USE_SSE ");
#endif
#if USE_SSE2
	   fprintf(stderr,"USE_SSE2 ");
#endif
#if USE_SSE3
	   fprintf(stderr,"USE_SSE3 ");
#endif
#if USE_SSSE3
	   fprintf(stderr,"USE_SSSE3 ");
#endif
#if USE_SSE41
	   fprintf(stderr,"USE_SSE4.1 ");
#endif
#elif _GNU_SOURCE || __APPLE__ || __FreeBSD__ //uje: shows only highest used streaming SIMD in stderr
#if USE_SSE && !(USE_SSE2 || USE_SSE3 || USE_SSSE3 || USE_SSE41 || USE_AVX)
	   fprintf(stderr,"SSE ");
#endif
#if USE_SSE2 && !(USE_SSE3 || USE_SSSE3 || USE_SSE41 || USE_AVX)
	   fprintf(stderr,"SSE2 ");
#endif
#if USE_SSE3 && !(USE_SSSE3 || USE_SSE41 || USE_AVX)
	   fprintf(stderr,"SSE3 ");
#endif
#if USE_SSSE3 && !(USE_SSE41 || USE_AVX)
	   fprintf(stderr,"SSSE3 ");
#endif
#if USE_SSE41 && !(USE_AVX)
	   fprintf(stderr,"SSE4.1 ");
#endif
#if USE_AVX
	   fprintf(stderr,"AVX ");
#endif
#endif // _GNU_SOURCE || __APPLE__ || __FreeBSD__
#if _WIN64
	   fprintf(stderr,"x64 ");
#elif _WIN32
	   fprintf(stderr,"x86 ");
#endif

#if !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
const char* szFeatures[] =
{
    "FPU ",
    "",
    "",
    "",
    "TSC ",
    "",
    "PAE ",
    "",
    "CMPXCHG8B ",
    "APIC ",
    "",
    "SYSENTER ",
    "MTRR ",
    "",
    "",
    "CMOV/CCMP ",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "MMX ",
    "FXSAVE/FXRSTOR ",
    "SSE ",
    "SSE2 ",
    "",
    "HT ",
    "",
    "",
    ""
};
    __cpuid(CPUInfo, 0);
    nIds = CPUInfo[0];
    memset(CPUString, 0, sizeof(CPUString));
    *((int*)CPUString) = CPUInfo[1];
    *((int*)(CPUString+4)) = CPUInfo[3];
    *((int*)(CPUString+8)) = CPUInfo[2];

    // Get the information associated with each valid Id
    for (i=0; i<=nIds; ++i)
    {
        __cpuid(CPUInfo, i);
   // Interpret CPU feature information.
        if  (i == 1)
        {
            nSteppingID = CPUInfo[0] & 0xf;
            nModel = (CPUInfo[0] >> 4) & 0xf;
            nFamily = (CPUInfo[0] >> 8) & 0xf;
            nProcessorType = (CPUInfo[0] >> 12) & 0x3;
            nExtendedmodel = (CPUInfo[0] >> 16) & 0xf;
            nExtendedfamily = (CPUInfo[0] >> 20) & 0xff;
            nBrandIndex = CPUInfo[1] & 0xff;
            nCLFLUSHcachelinesize = ((CPUInfo[1] >> 8) & 0xff) * 8;
            nAPICPhysicalID = (CPUInfo[1] >> 24) & 0xff;
            bSSE3NewInstructions = (CPUInfo[2] & 0x1) || false;
            bMONITOR_MWAIT = (CPUInfo[2] & 0x8) || false;
            bCPLQualifiedDebugStore = (CPUInfo[2] & 0x10) || false;
            bThermalMonitor2 = (CPUInfo[2] & 0x100) || false;
            nFeatureInfo = CPUInfo[3];
        }
    }

    // Calling __cpuid with 0x80000000 as the InfoType argument
    // gets the number of valid extended IDs.
    __cpuid(CPUInfo, 0x80000000);
    nExIds = CPUInfo[0];
    memset(CPUBrandString, 0, sizeof(CPUBrandString));

    // Get the information associated with each extended ID.
    for (i=0x80000000; i<=nExIds; ++i)
    {
        __cpuid(CPUInfo, i);
   // Interpret CPU brand string and cache information.
        if  (i == 0x80000002)
            memcpy(CPUBrandString, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000003)
            memcpy(CPUBrandString + 16, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000004)
            memcpy(CPUBrandString + 32, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000006)
        {
            nCacheLineSize = CPUInfo[2] & 0xff;
            nL2Associativity = (CPUInfo[2] >> 12) & 0xf;
            nCacheSizeK = (CPUInfo[2] >> 16) & 0xffff;
 		}
    }
    // Display all the information in user-friendly format.
   if  (nExIds >= 0x80000004)
	   fprintf( stderr, "\n%10s: %s \n"
            , "CPUID"
            , CPUBrandString);
#if USE_IPP
   int myfreq;
    ippGetCpuFreqMhz(&myfreq);
    fprintf(stderr,"%10s: %d x %d MHz ","Speed",ippGetNumCoresOnDie(), myfreq);
#endif
fprintf(stderr,"\n");
  if  (nExIds >= 0x80000006)
    {
		fprintf( stderr, "%10s: L1=%dK L2=%dK\n"
			    , "Cache"
				, nCacheLineSize
				, nCacheSizeK);
	}

  if  (nIds >= 1)
    {
		   if  (nFeatureInfo || bSSE3NewInstructions ||
             bMONITOR_MWAIT || bCPLQualifiedDebugStore ||
             bThermalMonitor2)
        {
  if  (bSSE3NewInstructions)
 //R: this info is misleading. If, for example, SSE3 app will be run on SSE4.1 capable CPU it will show only SSE3 level of features.
#if defined (USE_SSE41)
fprintf( stderr, "%10s: %s%s%s%s%s%s \n"
            , "Features used"
            , "MMX "
            , "SSE "
			, "SSE2 "
			, "SSE3 "
			, "SSSE3 "
            , "SSE4.1 \n");
#elif defined (USE_SSSE3)
fprintf( stderr, "%10s: %s%s%s%s%s \n"
            , "Features used"
            , "MMX "
            , "SSE "
			, "SSE2 "
			, "SSE3 "
            , "SSSE3 \n");
#elif defined (USE_SSE3)
fprintf( stderr, "%10s: %s%s%s%s \n"
            , "Features used"
            , "MMX "
            , "SSE "
			, "SSE2 "
            , "SSE3 \n");
#elif defined (USE_SSE2)
fprintf( stderr, "%10s: %s%s%s \n"
            , "Features used"
            , "MMX "
            , "SSE "
            , "SSE2 \n");
#elif defined (USE_SSE)
fprintf( stderr, "%10s: %s%s \n"
            , "Features used"
            , "MMX "
            , "SSE\n");
#elif defined (USE_MMX)
fprintf( stderr, "%10s: %s \n"
            , "Features used"
            , "MMX \n");
#endif

            i = 0;
            nIds = 1;
			fprintf(stderr,"\nCPU features: ");
			while (i < (sizeof(szFeatures)/sizeof(const char*)))
            {
                if  (nFeatureInfo & nIds)
                {
                    fprintf(stderr,szFeatures[i]);
                }
                nIds <<= 1;
                ++i;
            }
			if  (bSSE3NewInstructions)fprintf(stderr,"SSE3 ");
#ifdef USE_IPP
			IppCpuType cpu_type=ippGetCpuType ();
			if(cpu_type==ippCpuSSE41)fprintf(stderr,"SSE4.1 ");
			else if(cpu_type==ippCpuSSE42)fprintf(stderr,"SSE4.2 ");
			if(cpu_type==ippCpuEM64T || cpu_type==ippCpuX8664)fprintf(stderr,"x64 ");
			fprintf(stderr,"\nCPU type 0x%x ",cpu_type);
#endif
			fprintf(stderr,"\n");
        }
   }
#endif // ! _GNU_SOURCE || __APPLE__ || __FreeBSD__
} //intrinsic_cpuid()

#if _GNU_SOURCE || __APPLE__ || __FreeBSD__
//uje: Lets see what OS we are at :
void give_system_info(){
    utsname *buf = new utsname();
    uname(buf);
    const char* sysnam = buf->sysname;
    const char* releas = buf->release;
    const char* machin = buf->machine;
    fprintf(stderr, "\n System: %s ", sysnam);
    fprintf(stderr, " %s ", machin);
    fprintf(stderr, " Kernel: %s\n", releas);
    delete(buf);
}
/* struct utsname {
               char sysname[];
               char nodename[];
               char release[];
               char version[];
               char machine[];
           #ifdef _GNU_SOURCE
               char domainname[];
           #endif
           };*/

#if !__APPLE__
//uje: Lets read some CPU data from /proc/* if it is available.
void give_cpu_data() {

    std::string ch;
#if _GNU_SOURCE
    const std::string cpudata="/proc/cpuinfo";
#elif __FreeBSD__
    const std::string cpudata="/compat/linux/proc/cpuinfo";
#else
    goto end;
#endif
    //uje: This is definitly works (currently) on linux only,
    //because it is OS specific since kernel 2.6.22. Older
    //kernels since 2.6.18 could have been patched with /proc also.
    //FreeBSD is moving to /proc later.
    //Apple is using their own (poss. system-profiler). => Todo!
    short num=0;
    std::string MN1, SP, L1, L2;

    std::ifstream give(cpudata.c_str());
    if (!give) {
#if _GNU_SOURCE
    fprintf(stderr, "File %s not found. Continuing ...\n",(cpudata.c_str()));
#endif
		goto end;
	}

    while (!give.eof()) {
        getline(give,ch);
        if (ch.find_first_not_of("processor")== 9) {
            num++;
        } else if (ch.find_first_not_of("model name")== 10) {
            ch.replace(ch.find("model name"),13,"");
            MN1=ch;
        } else if (ch.find_first_not_of("cpu MHz")== 7) {
            ch.replace(ch.find("cpu MHz"),10, "");
            SP=ch;
#if _GNU_SOURCE
        } else if (ch.find_first_not_of("cache size")== 10) {
            ch.replace(ch.find("cache size"),13, "");
            L2=ch;
        } else if (ch.find_first_not_of("clflush size")== 12) {
            ch.replace(ch.find("clflush size"),15, "");
            L1=ch;
#endif
        }
    }
    fprintf(stderr," CPU   : %s\n", MN1.c_str());
    fprintf(stderr," %d core(s), Speed : %s MHz\n", num, SP.c_str());
#if _GNU_SOURCE
    fprintf(stderr," L1 : %s KB, Cache : %s\n", L1.c_str(), L2.c_str());
#endif

end:
	if (!give) fprintf(stderr,"\n");
}
#endif
#endif
